@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -204,6 +204,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			{
 				pool_query_context_destroy(query_context);
 				pool_set_skip_reading_from_backends();
+				pool_memory = old_context;
 				return POOL_CONTINUE;
 			}
 		}
@@ -222,6 +223,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			if (parallel)
 			{
 				pool_query_context_destroy(query_context);
+				pool_memory = old_context;
 				return status;
 			}
 		}
@@ -275,6 +277,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 			pool_query_context_destroy(query_context);
 			pool_set_skip_reading_from_backends();
+			pool_memory = old_context;
 			return POOL_CONTINUE;
 		}
 
@@ -694,8 +697,6 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			return POOL_END;
 		}
 
-		session_context->uncompleted_message = msg;
-
 		/*
 		 * Decide where to send query
 		 */
@@ -795,6 +796,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 	}
 
+	/*
+	 * need to set uncompleted_message after calling ReadyForQuery(),
+	 * because the function destroys query context of uncompleted_message.
+	 */
+	session_context->uncompleted_message = msg;
+
 	/*
 	 * Cannot call free_parser() here. Since "string" might be allocated in parser context.
 	 * free_parser();
@@ -1383,10 +1390,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		pool_unset_query_in_progress();
 	}
 
-	if (!pool_is_doing_extended_query_message())
+	if (session_context->query_context && !pool_is_doing_extended_query_message())
 	{
 		if (!(node && IsA(node, PrepareStmt)))
-			pool_query_context_destroy(pool_get_session_context()->query_context);
+			pool_query_context_destroy(session_context->query_context);
 	}
 
 	sp = MASTER_CONNECTION(backend)->sp;
@@ -1724,6 +1731,20 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 			p1 = p2;
 			len1 = strlen(p2) + 1;
 		}
+		/*
+		 * If PREPARE or extended query protocol commands caused error,
+		 * remove the temporary saved message.
+		 */
+		else
+		{
+			if (session_context->uncompleted_message)
+			{
+				pool_add_sent_message(session_context->uncompleted_message);
+				pool_remove_sent_message(session_context->uncompleted_message->kind,
+										 session_context->uncompleted_message->name);
+				session_context->uncompleted_message = NULL;
+			}
+		}
 
 		pool_write(frontend, "C", 1);
 		sendlen = htonl(len1+4);
@@ -1746,7 +1767,6 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS ret;
-	POOL_SESSION_CONTEXT *session_context;
 
 	ret = SimpleForwardToFrontend('E', frontend, backend);
 	if (ret != POOL_CONTINUE)
@@ -1811,21 +1831,6 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 	}
 #endif
 
-	/* An error occurred with PREPARE or DEALLOCATE command.
-	 * Free pending portal object.
-	 */
-	session_context = pool_get_session_context();
-	if (session_context)
-	{
-		if (session_context->uncompleted_message)
-		{
-			pool_add_sent_message(session_context->uncompleted_message);
-			pool_remove_sent_message(session_context->uncompleted_message->kind,
-									 session_context->uncompleted_message->name);
-			session_context->uncompleted_message = NULL;
-		}
-	}
-
 	return POOL_CONTINUE;
 }
 
@@ -1972,7 +1977,15 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		return POOL_END;
 
 	if (fkind != 'S' && pool_is_ignore_till_sync())
+	{
+		/*
+		 * Flag setting for calling ProcessBackendResponse()
+		 * in pool_process_query().
+		 */
+		if (!pool_is_query_in_progress())
+			pool_set_query_in_progress();
 		return POOL_CONTINUE;
+	}
 
 	pool_unset_doing_extended_query_message();
 
@@ -2097,6 +2110,8 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 	if (pool_is_ignore_till_sync())
 	{
+		if (pool_is_query_in_progress())
+			pool_unset_query_in_progress();
 		return POOL_CONTINUE;
 	}
 
